// Written by Syranide, me@syranide.com

if SERVER then return end

-- mode for showing grid only on USE, same for cursor
-- mode for showing one snap point only? (connect with above)

-- see if it possible to fix the wobbling when moving (doesn't wobble when you move normally)

local player =        LocalPlayer()
local target =        { active = false }
local snaptarget =    { active = false }
local snapkey =       false
local snaptime =      false
local snaplock =      false
local snapcursor =    false
local snapspawnmenu = false

local convars = {
	enabled      = 1,
	toggledelay  = 0,
	disableuse   = 0,
	allentities  = 0,
	alltools     = 0,
	enabletoggle = 0,
	lockdelay    = 0.5,
	distance     = 250,
	gridlimit    = 16,
	gridsize     = 8,
	gridalpha    = 0.4,
	gridoffset   = 0.5,
	boundingbox  = 1,
	revertaim    = 1,
	centerline   = 1,
}

---------

local function StringExplode(sep, str)
	local lines = {}
	
	while str and str ~= "" do
		local pos = string.find(str, sep)
		if pos == nil then
			table.insert(lines, str)
			break
		end
		
		if pos > 1 then
			table.insert(lines, string.sub(str, 1, pos - 1))				
		else
			table.insert(lines, "")
		end
		
		str = string.sub(str, pos + 1)
	end
	
	return lines
end

local function ToScreen(p)
	local s = p:ToScreen()
	return Vector(s.x, s.y, 0)
end

local function PointToScreen(pp)
	local nc = player:GetAimVector()
	local pc = player:GetShootPos() + nc * 8
	
	local sp = nc:DotProduct(pp - pc)

	if (sp > 0) then 
		local s = pp:ToScreen()
		return Vector(s.x, s.y, 0)
	else
		return
	end	
end

local function LineToScreen(p1, p2)
	local nc = player:GetAimVector()
	local pc = player:GetShootPos() + nc * 8

	local s1 = nc:DotProduct(p1 - pc)
	local s2 = nc:DotProduct(p2 - pc)
	
	if s1 > 0 and s2 > 0 then
		return ToScreen(p1), ToScreen(p2)
	elseif s1 <= 0 and s2 <= 0 then
		return
	else
		local pd = p2 - p1
		local pi = p1 + pd * ((pc:DotProduct(nc) - p1:DotProduct(nc)) / pd:DotProduct(nc))
		
		if s1 <= 0 then
			return ToScreen(pi), ToScreen(p2)
		elseif s2 <= 0 then
			return ToScreen(p1), ToScreen(pi)
		end
	end
end


local function DrawSurfaceLine(v1, v2)
	surface.DrawLine(v1.x, v1.y, v2.x, v2.y)
end

------------

for key,value in pairs(convars) do
	CreateClientConVar('snap_' .. key, value, true, false)
end

local modelsaveset = {}
local modeloffsets = {}

local function OnInitialize()
	local data = file.Read('smartsnap_settings.txt')
	if data then
		lines = StringExplode("\n", data)
		header = table.remove(lines, 1)
		if header == 'SMARTSNAP_SETTINGS' then
			for _,line in ipairs(lines) do
				local pos = string.find(line, '=')
				if pos then
					 local key = string.Trim(string.sub(line, 1, pos - 1))
					 local value = string.Trim(string.sub(line, pos + 1))
					 
					 if convars[key] then player:ConCommand('snap_' .. key .. ' "' .. value .. '"') end
				end
			end
		end
	end

	for _,filename in ipairs({'smartsnap_default_offsets.txt', 'smartsnap_custom_offsets.txt'}) do
		local file = file.Read(filename)
		if file then
			lines = StringExplode("\n", file)
			header = table.remove(lines, 1)
			if header == "SMARTSNAP_OFFSETS" then
				for _,line in ipairs(lines) do
					local pos = string.find(line, '=')
					if pos then
						 local key = string.Trim(string.sub(line, 1, pos - 1))
						 local value = string.Trim(string.sub(line, pos + 1))
						 
						 local c = StringExplode(",", value)
						 modeloffsets[key]= { tonumber(c[1]), tonumber(c[2]), tonumber(c[3]), tonumber(c[4]),  tonumber(c[5]),  tonumber(c[6])  }
					end
				end
				
				--Msg("SmartSnap: " .. filename .. " ... OK\n")
			else
				--Msg("SmartSnap: " .. filename .. " ... ERROR\n")
			end
		else
			--Msg("SmartSnap: " .. filename .. " ... FAILED\n")
		end
	end
end
	
local function OnShutDown()
	output = "SMARTSNAP_SETTINGS\n"
	
	for key,_ in pairs(convars) do
		output = output .. key .. '=' .. GetConVarString('snap_' .. key) .. "\n"
	end
	
	file.Write('smartsnap_settings.txt', output)
	
	output = file.Read('smartsnap_custom_offsets.txt')
	if output == nil then
		output = "SMARTSNAP_OFFSETS\n"
	end
	
	for model,_ in pairs(modelsaveset) do
		output = output .. model .. '=' .. table.concat(modeloffsets[model], ",") .. "\n"
	end
	
	file.Write('smartsnap_custom_offsets.txt', output)
end

local function VEX()
	if modeloffsets[target.entity:GetModel()] == nil then
		modeloffsets[target.entity:GetModel()] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 }
	end
	return modeloffsets[target.entity:GetModel()]
end

concommand.Add("snap_dev_gridoffset", function(player, command, arguments) if target.active == true then if #arguments >= 1 then local v = VEX() v[target.face] = v[target.face] + tonumber(arguments[1]) end end end)
concommand.Add("snap_dev_saveoffset", function(player, command, arguments) if target.active == true then local v = VEX() modelsaveset[target.entity:GetModel()] = true end end)

local function SnapToggleGrid()
	if (GetConVarNumber("snap_enabled") == 0) then
		player:ConCommand('snap_enabled 1')
	else
		player:ConCommand('snap_enabled 0')
	end
end

local function SnapPress()
	if (snaplock or snapcursor) then
		snaptime = false
	else
		local toggledelay = GetConVarNumber("snap_toggledelay")
		if (toggledelay > 0 and snaptime and snaptime + toggledelay > CurTime()) then
			SnapToggleGrid()
			snaptime = false
			snaplock = false
		else
			snaptime = CurTime()
		end
	end
	
	snapkey = target.active
	
	if (!snapcursor) then
		snaplock = false
	end
end

local function SnapRelease()
	snapkey = false
end

local function SnapLock()
	snaplock = !snaplock
end

local function OnSpawnMenu()
	snapspawnmenu = true
end

local function OnKeyPress(player, key)
	if (key == IN_USE and GetConVarNumber("snap_disableuse") == 0) then
		SnapPress()
	end
end

local function OnKeyRelease(player, key)
	if (key == IN_USE and GetConVarNumber("snap_disableuse") == 0) then
		SnapRelease()
	end
end

local function OnThink()
	if (vgui.CursorVisible()) then
		if (!snapcursor and snaplock) then
			snaptarget = table.Copy(target)
		end
		
		snaptime = false
		snapcursor = true
	else
		if (snapcursor and snaplock) then
			target = snaptarget
		end
		
		snapspawnmenu = false
		snapcursor = false
	end
	
	if (GetConVarNumber("snap_enabletoggle") != 0) then
		if (snapkey and snaptime and !snaplock) then
			if (CurTime() > snaptime + GetConVarNumber("snap_lockdelay")) then
				snaplock = true
				snaptime = false
			end
		end
	end
		
	local locked = target.locked and target.active
	target.locked = (snapkey or snaplock and !snapcursor) and target.active
	
	if (!target.locked and locked and GetConVarNumber("snap_revertaim") != 0) then
		if (snapcursor) then
			local screen = target.entity:LocalToWorld(target.vector):ToScreen()
			gui.SetMousePos(math.Round(screen.x), math.Round(screen.y))
		else
			local angles = (target.entity:LocalToWorld(target.vector) - player:GetShootPos()):Angle()
			player:SetEyeAngles(angles)
		end
	end
end

local function PlaneRayIntersect(o, v, p, e1, e2)
	local vp = v:Cross(e2)

	local d = e1:DotProduct(vp)
	if (d <= 0.0) then return nil end

	local vt = o - p
	local u = vt:DotProduct(vp)
	if (u < 0.0 or u > d) then return nil end

	local v = v:DotProduct(vt:Cross(e1))
	if (v < 0.0 or v > d) then return nil end

	local id = 1.0 / d
	return Vector(u * id, v * id, 0)
end

local function snap(X, Y, v)
	local lengthX = X:Length()
	local lengthY = Y:Length()
	local offsetX = GetConVarNumber("snap_gridoffset") / lengthX
	local offsetY = GetConVarNumber("snap_gridoffset") / lengthY
	local scaleX = 1 - offsetX * 2
	local scaleY = 1 - offsetY * 2
	local gridX = math.Clamp(math.floor(lengthX / (2 * GetConVarNumber("snap_gridsize")) * scaleX) * 2, 2, GetConVarNumber("snap_gridlimit"))
	local gridY = math.Clamp(math.floor(lengthY / (2 * GetConVarNumber("snap_gridsize")) * scaleY) * 2, 2, GetConVarNumber("snap_gridlimit"))
	
	BX = math.Clamp(math.Round((v.x - offsetX) / scaleX * gridX), 0, gridX)
	BY = math.Clamp(math.Round((v.y - offsetY) / scaleY * gridY), 0, gridY)
	
	return X * (offsetX + (BX / gridX) * scaleX) + Y * (offsetY + (BY / gridY) * scaleY)
end

local function DrawGridLines(O,X,Y,gridX,offsetX,scaleX,offsetY,scaleY,sign)
	for i = 0, gridX do
		local baseX = O + X * (offsetX + (i / gridX) * scaleX)
		local A = baseX + Y * (offsetY)
		local B = baseX + Y * (offsetY + scaleY)
		
		A, B = LineToScreen(A, B)
		
		if (A) then
			local normal = (A - B):Normalize()
			if math.abs(normal.x) < math.abs(normal.y) then b = Vector(0.5, 0, 0) else b = Vector(0, 0.5, 0) end
			if math.abs(normal.x) < 1 - math.abs(normal.y) then b = Vector(0, 0, 0) - b end
			DrawSurfaceLine(A + b * sign, B + b * sign)
			
			if (gridX / 2 == i and GetConVarNumber("snap_centerline") != 0) then
				DrawSurfaceLine(A + b * sign * -3, B + b * sign * -3)
			end
		end
	end
end

local function drawgrid(O, X, Y)
	local lengthX = X:Length()
	local lengthY = Y:Length()
	local offsetX = GetConVarNumber("snap_gridoffset") / lengthX
	local offsetY = GetConVarNumber("snap_gridoffset") / lengthY
	local scaleX = 1 - offsetX * 2
	local scaleY = 1 - offsetY * 2
	local gridX = math.Clamp(math.floor(lengthX / (2 * GetConVarNumber("snap_gridsize")) * scaleX) * 2, 2, GetConVarNumber("snap_gridlimit"))
	local gridY = math.Clamp(math.floor(lengthY / (2 * GetConVarNumber("snap_gridsize")) * scaleY) * 2, 2, GetConVarNumber("snap_gridlimit"))
	
	surface.SetDrawColor(0, 0, 0, math.Round(GetConVarNumber("snap_gridalpha") * 255))
	DrawGridLines(O, X, Y, gridX, offsetX, scaleX, offsetY, scaleY, -1)
	DrawGridLines(O, Y, X, gridY, offsetY, scaleY, offsetX, scaleX, -1)
	
	surface.SetDrawColor(255, 255, 255, math.Round(GetConVarNumber("snap_gridalpha") * 255))
	DrawGridLines(O, X, Y, gridX, offsetX, scaleX, offsetY, scaleY,  1)
	DrawGridLines(O, Y, X, gridY, offsetY, scaleY, offsetX, scaleX,  1)
end

local function drawcorner(C, CH)
	X = C + (CH - C):Normalize() * 5
	
	if (C:Distance(X) > C:Distance(CH)) then
		X = C - (C - CH) / 2
	end
	
	XX, CC = LineToScreen(X, C)
	
	if (XX) then
		surface.SetDrawColor(0, 0, 255, 192)
		DrawSurfaceLine(XX, CC)
	end
end

local function drawcorners(C, CH, CV, CS)
	drawcorner(C, CH)
	drawcorner(C, CV)
	drawcorner(C, CS)
end

local function drawmarker(marker, r, g, b)
	surface.SetDrawColor(0, 0, 0, 255)
	DrawSurfaceLine(Vector(marker.x - 2.5, marker.y - 2), Vector(marker.x + 2.5, marker.y + 3))
	DrawSurfaceLine(Vector(marker.x + 1.5, marker.y - 2), Vector(marker.x - 3.5, marker.y + 3))
	
	surface.SetDrawColor(r, g, b, 255)
	DrawSurfaceLine(Vector(marker.x - 1.5, marker.y - 2), Vector(marker.x + 3.5, marker.y + 3))
	DrawSurfaceLine(Vector(marker.x + 2.5, marker.y - 2), Vector(marker.x - 2.5, marker.y + 3))
end

local function OnPaintHUD()
	target.active = false
	
	if (GetConVarNumber("snap_enabled") == 0) then return end	
	if (!player:Alive() or player:InVehicle()) then return end
	
	if (target.locked) then
		if (!target.entity:IsValid()) then return end
	else
		local trace = player:GetEyeTrace()
		if (!trace.HitNonWorld) then return end
		
		local entity = trace.Entity
		if (entity == nil) then return end
			
		if (entity:GetClass() != "prop_physics" and GetConVarNumber("snap_allentities") == 0) then return end
		if (!player:GetActiveWeapon():IsValid()) then return end
		if (player:GetActiveWeapon():GetClass() == "weapon_physgun") then return end
		if (player:GetActiveWeapon():GetClass() != "gmod_tool" and GetConVarNumber("snap_alltools") == 0) then return end
		
		target.entity = entity
	end
	
	local offsets
	if modeloffsets[target.entity:GetModel()] then
		offsets = modeloffsets[target.entity:GetModel()]
	else
		offsets = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 }
	end
	
	local obbmax = target.entity:OBBMaxs()
	local obbmin = target.entity:OBBMins()
	
	local obvsnap = {
		lsw = target.entity:LocalToWorld(Vector(obbmin.x, obbmin.y, obbmin.z)),
		lse = target.entity:LocalToWorld(Vector(obbmax.x, obbmin.y, obbmin.z)),
		lnw = target.entity:LocalToWorld(Vector(obbmin.x, obbmax.y, obbmin.z)),
		lne = target.entity:LocalToWorld(Vector(obbmax.x, obbmax.y, obbmin.z)),
		usw = target.entity:LocalToWorld(Vector(obbmin.x, obbmin.y, obbmax.z)),
		use = target.entity:LocalToWorld(Vector(obbmax.x, obbmin.y, obbmax.z)),
		unw = target.entity:LocalToWorld(Vector(obbmin.x, obbmax.y, obbmax.z)),
		une = target.entity:LocalToWorld(Vector(obbmax.x, obbmax.y, obbmax.z)),
	}
	
	local obbmax = obbmax - Vector(offsets[5], offsets[3], offsets[1])
	local obbmin = obbmin + Vector(offsets[6], offsets[4], offsets[2])
	
	local obvgrid = {
		lsw = target.entity:LocalToWorld(Vector(obbmin.x, obbmin.y, obbmin.z)),
		lse = target.entity:LocalToWorld(Vector(obbmax.x, obbmin.y, obbmin.z)),
		lnw = target.entity:LocalToWorld(Vector(obbmin.x, obbmax.y, obbmin.z)),
		lne = target.entity:LocalToWorld(Vector(obbmax.x, obbmax.y, obbmin.z)),
		usw = target.entity:LocalToWorld(Vector(obbmin.x, obbmin.y, obbmax.z)),
		use = target.entity:LocalToWorld(Vector(obbmax.x, obbmin.y, obbmax.z)),
		unw = target.entity:LocalToWorld(Vector(obbmin.x, obbmax.y, obbmax.z)),
		une = target.entity:LocalToWorld(Vector(obbmax.x, obbmax.y, obbmax.z)),
	}
	
	local shootPoint = player:GetShootPos()
	local aimvector = player:GetCursorAimVector()
	
	if (!target.locked) then
		-- should improve this by expanding the bounding box or something instead!
		-- create a larger bounding box and then planes for each side, and check distance from the plane
		local distance = (player:GetPos() - target.entity:GetPos()):Length() - (obvgrid.unw - obvgrid.lse):Length()
		if (distance > GetConVarNumber("snap_distance")) then return end		
		
		local faces = {
			{ obvgrid.unw, obvgrid.usw - obvgrid.unw, obvgrid.une - obvgrid.unw },
			{ obvgrid.lsw, obvgrid.lnw - obvgrid.lsw, obvgrid.lse - obvgrid.lsw },
			{ obvgrid.unw, obvgrid.une - obvgrid.unw, obvgrid.lnw - obvgrid.unw },
			{ obvgrid.usw, obvgrid.lsw - obvgrid.usw, obvgrid.use - obvgrid.usw },
			{ obvgrid.une, obvgrid.use - obvgrid.une, obvgrid.lne - obvgrid.une },
			{ obvgrid.unw, obvgrid.lnw - obvgrid.unw, obvgrid.usw - obvgrid.unw },
		}
		
		for face,vertices in ipairs(faces) do
			intersection = PlaneRayIntersect(shootPoint, aimvector, vertices[1], vertices[2], vertices[3])
			if (intersection) then
				target.face = face
				break
			end
		end
		
		if intersection == nil then return end
	end
	
	if target.face == 1 then
		vectorOrigin = obvgrid.unw
		vectorX = obvgrid.usw - obvgrid.unw
		vectorY = obvgrid.une - obvgrid.unw
		vectorNormal = obvgrid.unw - obvgrid.lnw
		
		vectorLength = obvsnap.lnw - obvsnap.unw
		vectorOffset = Vector(0, 0, -offsets[1])
	elseif target.face == 2 then
		vectorOrigin = obvgrid.lsw
		vectorX = obvgrid.lnw - obvgrid.lsw
		vectorY = obvgrid.lse - obvgrid.lsw
		vectorNormal = obvgrid.lsw - obvgrid.usw
		
		vectorLength = obvsnap.usw - obvsnap.lsw
		vectorOffset = Vector(0, 0, offsets[2])
	elseif target.face == 3 then
		vectorOrigin = obvgrid.unw
		vectorX = obvgrid.une - obvgrid.unw
		vectorY = obvgrid.lnw - obvgrid.unw
		vectorNormal = obvgrid.unw - obvgrid.usw
		
		vectorLength = obvsnap.usw - obvsnap.unw
		vectorOffset = Vector(0,-offsets[3], 0)
	elseif target.face == 4 then
		vectorOrigin = obvgrid.usw
		vectorX = obvgrid.lsw - obvgrid.usw
		vectorY = obvgrid.use - obvgrid.usw
		vectorNormal = obvgrid.usw - obvgrid.unw
		
		vectorLength = obvsnap.unw - obvsnap.usw
		vectorOffset = Vector(0, offsets[4], 0)
	elseif target.face == 5 then
		vectorOrigin = obvgrid.une
		vectorX = obvgrid.use - obvgrid.une
		vectorY = obvgrid.lne - obvgrid.une
		vectorNormal = obvgrid.une - obvgrid.unw
		
		vectorLength = obvsnap.unw - obvsnap.une
		vectorOffset = Vector(-offsets[5], 0, 0)
	elseif target.face == 6 then
		vectorOrigin = obvgrid.unw
		vectorX = obvgrid.lnw - obvgrid.unw
		vectorY = obvgrid.usw - obvgrid.unw
		vectorNormal = obvgrid.unw - obvgrid.une
		
		vectorLength = obvsnap.une - obvsnap.unw
		vectorOffset = Vector(offsets[6], 0, 0)
	end
	
	vectorNormal:Normalize()
	
	if (GetConVarNumber("snap_boundingbox") != 0) then
		drawcorners(obvgrid.unw, obvgrid.lnw, obvgrid.usw, obvgrid.une)
		drawcorners(obvgrid.une, obvgrid.lne, obvgrid.use, obvgrid.unw)
		drawcorners(obvgrid.lnw, obvgrid.unw, obvgrid.lsw, obvgrid.lne)
		drawcorners(obvgrid.lne, obvgrid.une, obvgrid.lse, obvgrid.lnw)
		drawcorners(obvgrid.usw, obvgrid.lsw, obvgrid.unw, obvgrid.use)
		drawcorners(obvgrid.use, obvgrid.lse, obvgrid.une, obvgrid.usw)
		drawcorners(obvgrid.lsw, obvgrid.usw, obvgrid.lnw, obvgrid.lse)
		drawcorners(obvgrid.lse, obvgrid.use, obvgrid.lne, obvgrid.lsw)
	end
	
	if (!target.locked) then
		vectorGrid = vectorOrigin + snap(vectorX, vectorY, intersection)
		
		local trace = util.TraceLine({
			start  = target.entity:LocalToWorld(target.entity:WorldToLocal(vectorGrid) - vectorOffset) + vectorNormal * 0.01,
			endpos = vectorGrid + vectorLength,
		})
		
		vectorSnap = trace.HitPos
		target.offset = target.entity:WorldToLocal(vectorSnap)
		target.vector = target.entity:WorldToLocal(vectorGrid)
		
		/*if (trace.Entity == nil or !trace.Entity:IsValid()) then
			trace = nil
		elseif (trace.Entity != target.entity) then
			trace = nil
		elseif (trace.HitPos == trace.StartPos) then
			trace = nil
		elseif ((trace.HitPos - trace.StartPos):Length() > (vectorLength):Length() / 2) then
			trace = nil
		end
		
		if (trace == nil) then
			trace = util.TraceLine({
				start  = target.entity:LocalToWorld(target.entity:WorldToLocal(vectorGrid) - vectorOffset) + vectorNormal * 0.1 + vectorY * 0.1 + vectorX * 0.1,
				endpos = vectorGrid + vectorLength,
			})
		end*/
		
		target.error = true

		if (trace.Entity == nil or !trace.Entity:IsValid()) then
			snaperror = -1
		elseif (trace.Entity != target.entity) then
			snaperror = -2
		elseif (trace.HitPos == trace.StartPos) then
			snaperror = -2
		else
			snaperror = (player:GetEyeTrace().HitPos - trace.HitPos):Length()
			target.error = false
			
			if ((vectorSnap - vectorGrid):Length() > 0.5) then
				local marker = PointToScreen(vectorSnap)
				
				if (marker) then
					drawmarker(marker, 255, 255, 255)
				end
			end
		end
	else
		vectorGrid = target.entity:LocalToWorld(target.vector)
		vectorSnap = target.entity:LocalToWorld(target.offset)
	
		snaperror = (player:GetEyeTrace().HitPos - vectorSnap):Length()
		
		local marker = PointToScreen(vectorSnap)
		
		if (marker) then
			if (target.error == true) then
				snaperror = -2
				drawmarker(marker, 0, 255, 255)
			elseif (snaperror < 0.001) then
				drawmarker(marker, 0, 255, 0)
			elseif (snaperror < 0.1) then
				drawmarker(marker, 255, 255, 0)
			else
				drawmarker(marker, 255, 0, 0)
			end
		end
	end
	
	drawgrid(vectorOrigin, vectorX, vectorY)

	target.active = true
	
	local cursor = PointToScreen(vectorGrid)
	
	if (cursor) then
		if (snaperror == -1) then
			target.active = false
			drawmarker(cursor, 0, 255, 255)
		elseif (snaperror == -2) then
			drawmarker(cursor, 255, 0, 255)
		elseif (snaperror < 0.001) then
			drawmarker(cursor, 0, 255, 0)
		elseif (snaperror < 0.1) then
			drawmarker(cursor, 255, 255, 0)
		else
			drawmarker(cursor, 255, 0, 0)
		end
	end
end

local function OnSnapView(player, origin, angles, fov)
	local targetvalid = target.active and target.locked and target.entity:IsValid()
	local snaptargetvalid = snaptarget.active and snaptarget.locked and snaptarget.entity:IsValid()

	if (snapcursor and !snapspawnmenu and targetvalid) then
		local screen = target.entity:LocalToWorld(target.offset):ToScreen()
		gui.SetMousePos(math.Round(screen.x), math.Round(screen.y))
	end
	
	if (!snapcursor and targetvalid) then
		local angles = (target.entity:LocalToWorld(target.offset) - player:GetShootPos()):Angle()
		return {angles = angles}
	elseif (snaplock and snaptargetvalid) then	
		local angles = (snaptarget.entity:LocalToWorld(snaptarget.offset) - player:GetShootPos()):Angle()
		return {angles = angles}
	end
end

local function OnSnapAim(user)
	local targetvalid = target.active and target.locked and target.entity:IsValid()
	local snaptargetvalid = snaptarget.active and snaptarget.locked and snaptarget.entity:IsValid()

	if (!snapcursor and targetvalid) then
		local angles = (target.entity:LocalToWorld(target.offset) - player:GetShootPos()):Angle()
		user:SetViewAngles(angles)
	elseif (snaplock and snaptargetvalid) then	
		local angles = (snaptarget.entity:LocalToWorld(snaptarget.offset) - player:GetShootPos()):Angle()
		user:SetViewAngles(angles)
	end
end


OnInitialize()


concommand.Add("+snap", SnapPress)
concommand.Add("-snap", SnapRelease)
concommand.Add("snaplock", SnapLock)
concommand.Add("snaptogglegrid", SnapToggleGrid)


hook.Add("SpawnMenuOpen", "SmartsnapSpawnMenu", OnSpawnMenu)

hook.Add("Think", "SmartsnapThink", OnThink)
hook.Add("ShutDown", "SmartsnapShutDown", OnShutDown)

hook.Add("KeyPress", "SmartsnapKeyPress", OnKeyPress)
hook.Add("KeyRelease", "SmartsnapKeyRelease", OnKeyRelease)

hook.Add("CreateMove", "SmartsnapSnap", OnSnapAim)
hook.Add("CalcView", "SmartsnapSnapView", OnSnapView)

hook.Add("SpawnMenuOpen", "SmartsnapSpawnMenu", OnSpawnMenu)

hook.Add("HUDPaintBackground", "SmartsnapPaintHUD", OnPaintHUD)
